@@ -1,8 +1,13 @@
 use crate::ast::{Enum, Field, Input, Struct};
+use crate::attr::Trait;
+use crate::generics::InferredBounds;
 use proc_macro2::TokenStream;
 use quote::{format_ident, quote, quote_spanned, ToTokens};
+use std::collections::BTreeSet as Set;
 use syn::spanned::Spanned;
-use syn::{Data, DeriveInput, Member, PathArguments, Result, Type, Visibility};
+use syn::{
+    Data, DeriveInput, GenericArgument, Member, PathArguments, Result, Token, Type, Visibility,
+};
 
 pub fn derive(node: &DeriveInput) -> Result<TokenStream> {
     let input = Input::from_syn(node)?;
@@ -16,14 +21,23 @@ pub fn derive(node: &DeriveInput) -> Result<TokenStream> {
 fn impl_struct(input: Struct) -> TokenStream {
     let ty = &input.ident;
     let (impl_generics, ty_generics, where_clause) = input.generics.split_for_impl();
+    let mut error_inferred_bounds = InferredBounds::new();
 
     let source_body = if input.attrs.transparent.is_some() {
-        let only_field = &input.fields[0].member;
+        let only_field = &input.fields[0];
+        if only_field.contains_generic {
+            error_inferred_bounds.insert(only_field.ty, quote!(std::error::Error));
+        }
+        let member = &only_field.member;
         Some(quote! {
-            std::error::Error::source(self.#only_field.as_dyn_error())
+            std::error::Error::source(self.#member.as_dyn_error())
         })
     } else if let Some(source_field) = input.source_field() {
         let source = &source_field.member;
+        if source_field.contains_generic {
+            let ty = unoptional_type(source_field.ty);
+            error_inferred_bounds.insert(ty, quote!(std::error::Error + 'static));
+        }
         let asref = if type_is_option(source_field.ty) {
             Some(quote_spanned!(source.span()=> .as_ref()?))
         } else {
@@ -39,66 +53,78 @@ fn impl_struct(input: Struct) -> TokenStream {
     let source_method = source_body.map(|body| {
         quote! {
             fn source(&self) -> std::option::Option<&(dyn std::error::Error + 'static)> {
-                use thiserror::private::AsDynError;
+                use thiserror::__private::AsDynError;
                 #body
             }
         }
     });
 
-    let backtrace_method = input.backtrace_field().map(|backtrace_field| {
+    let provide_method = input.backtrace_field().map(|backtrace_field| {
+        let demand = quote!(demand);
         let backtrace = &backtrace_field.member;
         let body = if let Some(source_field) = input.source_field() {
             let source = &source_field.member;
-            let source_backtrace = if type_is_option(source_field.ty) {
+            let source_provide = if type_is_option(source_field.ty) {
                 quote_spanned! {source.span()=>
-                    self.#source.as_ref().and_then(|source| source.as_dyn_error().backtrace())
+                    if let std::option::Option::Some(source) = &self.#source {
+                        source.provide(#demand);
+                    }
                 }
             } else {
                 quote_spanned! {source.span()=>
-                    self.#source.as_dyn_error().backtrace()
+                    self.#source.provide(#demand);
                 }
             };
-            let combinator = if source == backtrace {
-                source_backtrace
+            let self_provide = if source == backtrace {
+                None
             } else if type_is_option(backtrace_field.ty) {
-                quote! {
-                    #source_backtrace.or(self.#backtrace.as_ref())
-                }
+                Some(quote! {
+                    if let std::option::Option::Some(backtrace) = &self.#backtrace {
+                        #demand.provide_ref::<std::backtrace::Backtrace>(backtrace);
+                    }
+                })
             } else {
-                quote! {
-                    std::option::Option::Some(#source_backtrace.unwrap_or(&self.#backtrace))
-                }
+                Some(quote! {
+                    #demand.provide_ref::<std::backtrace::Backtrace>(&self.#backtrace);
+                })
             };
             quote! {
-                use thiserror::private::AsDynError;
-                #combinator
+                #[allow(unused_imports)]
+                use std::error::Error as _;
+                #source_provide
+                #self_provide
             }
         } else if type_is_option(backtrace_field.ty) {
             quote! {
-                self.#backtrace.as_ref()
+                if let std::option::Option::Some(backtrace) = &self.#backtrace {
+                    #demand.provide_ref::<std::backtrace::Backtrace>(backtrace);
+                }
             }
         } else {
             quote! {
-                std::option::Option::Some(&self.#backtrace)
+                #demand.provide_ref::<std::backtrace::Backtrace>(&self.#backtrace);
             }
         };
         quote! {
-            fn backtrace(&self) -> std::option::Option<&std::backtrace::Backtrace> {
+            fn provide<'_demand>(&'_demand self, #demand: &mut std::any::Demand<'_demand>) {
                 #body
             }
         }
     });
 
+    let mut display_implied_bounds = Set::new();
     let display_body = if input.attrs.transparent.is_some() {
         let only_field = &input.fields[0].member;
+        display_implied_bounds.insert((0, Trait::Display));
         Some(quote! {
             std::fmt::Display::fmt(&self.#only_field, __formatter)
         })
     } else if let Some(display) = &input.attrs.display {
+        display_implied_bounds = display.implied_bounds.clone();
         let use_as_display = if display.has_bonus_display {
             Some(quote! {
                 #[allow(unused_imports)]
-                use thiserror::private::{DisplayAsDisplay, PathAsDisplay};
+                use thiserror::__private::{DisplayAsDisplay, PathAsDisplay};
             })
         } else {
             None
@@ -114,14 +140,18 @@ fn impl_struct(input: Struct) -> TokenStream {
         None
     };
     let display_impl = display_body.map(|body| {
+        let mut display_inferred_bounds = InferredBounds::new();
+        for (field, bound) in display_implied_bounds {
+            let field = &input.fields[field];
+            if field.contains_generic {
+                display_inferred_bounds.insert(field.ty, bound);
+            }
+        }
+        let display_where_clause = display_inferred_bounds.augment_where_clause(input.generics);
         quote! {
             #[allow(unused_qualifications)]
-            impl #impl_generics std::fmt::Display for #ty #ty_generics #where_clause {
-                #[allow(
-                    // Clippy bug: https://github.com/rust-lang/rust-clippy/issues/7422
-                    clippy::nonstandard_macro_braces,
-                    clippy::used_underscore_binding,
-                )]
+            impl #impl_generics std::fmt::Display for #ty #ty_generics #display_where_clause {
+                #[allow(clippy::used_underscore_binding)]
                 fn fmt(&self, __formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
                     #body
                 }
@@ -131,7 +161,7 @@ fn impl_struct(input: Struct) -> TokenStream {
 
     let from_impl = input.from_field().map(|from_field| {
         let backtrace_field = input.distinct_backtrace_field();
-        let from = from_field.ty;
+        let from = unoptional_type(from_field.ty);
         let body = from_initializer(from_field, backtrace_field);
         quote! {
             #[allow(unused_qualifications)]
@@ -145,12 +175,18 @@ fn impl_struct(input: Struct) -> TokenStream {
     });
 
     let error_trait = spanned_error_trait(input.original);
+    if input.generics.type_params().next().is_some() {
+        let self_token = <Token![Self]>::default();
+        error_inferred_bounds.insert(self_token, Trait::Debug);
+        error_inferred_bounds.insert(self_token, Trait::Display);
+    }
+    let error_where_clause = error_inferred_bounds.augment_where_clause(input.generics);
 
     quote! {
         #[allow(unused_qualifications)]
-        impl #impl_generics #error_trait for #ty #ty_generics #where_clause {
+        impl #impl_generics #error_trait for #ty #ty_generics #error_where_clause {
             #source_method
-            #backtrace_method
+            #provide_method
         }
         #display_impl
         #from_impl
@@ -160,18 +196,27 @@ fn impl_struct(input: Struct) -> TokenStream {
 fn impl_enum(input: Enum) -> TokenStream {
     let ty = &input.ident;
     let (impl_generics, ty_generics, where_clause) = input.generics.split_for_impl();
+    let mut error_inferred_bounds = InferredBounds::new();
 
     let source_method = if input.has_source() {
         let arms = input.variants.iter().map(|variant| {
             let ident = &variant.ident;
             if variant.attrs.transparent.is_some() {
-                let only_field = &variant.fields[0].member;
+                let only_field = &variant.fields[0];
+                if only_field.contains_generic {
+                    error_inferred_bounds.insert(only_field.ty, quote!(std::error::Error));
+                }
+                let member = &only_field.member;
                 let source = quote!(std::error::Error::source(transparent.as_dyn_error()));
                 quote! {
-                    #ty::#ident {#only_field: transparent} => #source,
+                    #ty::#ident {#member: transparent} => #source,
                 }
             } else if let Some(source_field) = variant.source_field() {
                 let source = &source_field.member;
+                if source_field.contains_generic {
+                    let ty = unoptional_type(source_field.ty);
+                    error_inferred_bounds.insert(ty, quote!(std::error::Error + 'static));
+                }
                 let asref = if type_is_option(source_field.ty) {
                     Some(quote_spanned!(source.span()=> .as_ref()?))
                 } else {
@@ -190,7 +235,7 @@ fn impl_enum(input: Enum) -> TokenStream {
         });
         Some(quote! {
             fn source(&self) -> std::option::Option<&(dyn std::error::Error + 'static)> {
-                use thiserror::private::AsDynError;
+                use thiserror::__private::AsDynError;
                 #[allow(deprecated)]
                 match self {
                     #(#arms)*
@@ -201,7 +246,8 @@ fn impl_enum(input: Enum) -> TokenStream {
         None
     };
 
-    let backtrace_method = if input.has_backtrace() {
+    let provide_method = if input.has_backtrace() {
+        let demand = quote!(demand);
         let arms = input.variants.iter().map(|variant| {
             let ident = &variant.ident;
             match (variant.backtrace_field(), variant.source_field()) {
@@ -211,22 +257,26 @@ fn impl_enum(input: Enum) -> TokenStream {
                     let backtrace = &backtrace_field.member;
                     let source = &source_field.member;
                     let varsource = quote!(source);
-                    let source_backtrace = if type_is_option(source_field.ty) {
+                    let source_provide = if type_is_option(source_field.ty) {
                         quote_spanned! {source.span()=>
-                            #varsource.as_ref().and_then(|source| source.as_dyn_error().backtrace())
+                            if let std::option::Option::Some(source) = #varsource {
+                                source.provide(#demand);
+                            }
                         }
                     } else {
                         quote_spanned! {source.span()=>
-                            #varsource.as_dyn_error().backtrace()
+                            #varsource.provide(#demand);
                         }
                     };
-                    let combinator = if type_is_option(backtrace_field.ty) {
+                    let self_provide = if type_is_option(backtrace_field.ty) {
                         quote! {
-                            #source_backtrace.or(backtrace.as_ref())
+                            if let std::option::Option::Some(backtrace) = backtrace {
+                                #demand.provide_ref::<std::backtrace::Backtrace>(backtrace);
+                            }
                         }
                     } else {
                         quote! {
-                            std::option::Option::Some(#source_backtrace.unwrap_or(backtrace))
+                            #demand.provide_ref::<std::backtrace::Backtrace>(backtrace);
                         }
                     };
                     quote! {
@@ -235,8 +285,10 @@ fn impl_enum(input: Enum) -> TokenStream {
                             #source: #varsource,
                             ..
                         } => {
-                            use thiserror::private::AsDynError;
-                            #combinator
+                            #[allow(unused_imports)]
+                            use std::error::Error as _;
+                            #source_provide
+                            #self_provide
                         }
                     }
                 }
@@ -245,40 +297,51 @@ fn impl_enum(input: Enum) -> TokenStream {
                 {
                     let backtrace = &backtrace_field.member;
                     let varsource = quote!(source);
-                    let source_backtrace = if type_is_option(source_field.ty) {
+                    let source_provide = if type_is_option(source_field.ty) {
                         quote_spanned! {backtrace.span()=>
-                            #varsource.as_ref().and_then(|source| source.as_dyn_error().backtrace())
+                            if let std::option::Option::Some(source) = #varsource {
+                                source.provide(#demand);
+                            }
                         }
                     } else {
                         quote_spanned! {backtrace.span()=>
-                            #varsource.as_dyn_error().backtrace()
+                            #varsource.provide(#demand);
                         }
                     };
                     quote! {
                         #ty::#ident {#backtrace: #varsource, ..} => {
-                            use thiserror::private::AsDynError;
-                            #source_backtrace
+                            #[allow(unused_imports)]
+                            use std::error::Error as _;
+                            #source_provide
                         }
                     }
                 }
                 (Some(backtrace_field), _) => {
                     let backtrace = &backtrace_field.member;
                     let body = if type_is_option(backtrace_field.ty) {
-                        quote!(backtrace.as_ref())
+                        quote! {
+                            if let std::option::Option::Some(backtrace) = backtrace {
+                                #demand.provide_ref::<std::backtrace::Backtrace>(backtrace);
+                            }
+                        }
                     } else {
-                        quote!(std::option::Option::Some(backtrace))
+                        quote! {
+                            #demand.provide_ref::<std::backtrace::Backtrace>(backtrace);
+                        }
                     };
                     quote! {
-                        #ty::#ident {#backtrace: backtrace, ..} => #body,
+                        #ty::#ident {#backtrace: backtrace, ..} => {
+                            #body
+                        }
                     }
                 }
                 (None, _) => quote! {
-                    #ty::#ident {..} => std::option::Option::None,
+                    #ty::#ident {..} => {}
                 },
             }
         });
         Some(quote! {
-            fn backtrace(&self) -> std::option::Option<&std::backtrace::Backtrace> {
+            fn provide<'_demand>(&'_demand self, #demand: &mut std::any::Demand<'_demand>) {
                 #[allow(deprecated)]
                 match self {
                     #(#arms)*
@@ -290,6 +353,7 @@ fn impl_enum(input: Enum) -> TokenStream {
     };
 
     let display_impl = if input.has_display() {
+        let mut display_inferred_bounds = InferredBounds::new();
         let use_as_display = if input.variants.iter().any(|v| {
             v.attrs
                 .display
@@ -298,7 +362,7 @@ fn impl_enum(input: Enum) -> TokenStream {
         }) {
             Some(quote! {
                 #[allow(unused_imports)]
-                use thiserror::private::{DisplayAsDisplay, PathAsDisplay};
+                use thiserror::__private::{DisplayAsDisplay, PathAsDisplay};
             })
         } else {
             None
@@ -309,34 +373,41 @@ fn impl_enum(input: Enum) -> TokenStream {
             None
         };
         let arms = input.variants.iter().map(|variant| {
+            let mut display_implied_bounds = Set::new();
             let display = match &variant.attrs.display {
-                Some(display) => display.to_token_stream(),
+                Some(display) => {
+                    display_implied_bounds = display.implied_bounds.clone();
+                    display.to_token_stream()
+                }
                 None => {
                     let only_field = match &variant.fields[0].member {
                         Member::Named(ident) => ident.clone(),
                         Member::Unnamed(index) => format_ident!("_{}", index),
                     };
+                    display_implied_bounds.insert((0, Trait::Display));
                     quote!(std::fmt::Display::fmt(#only_field, __formatter))
                 }
             };
+            for (field, bound) in display_implied_bounds {
+                let field = &variant.fields[field];
+                if field.contains_generic {
+                    display_inferred_bounds.insert(field.ty, bound);
+                }
+            }
             let ident = &variant.ident;
             let pat = fields_pat(&variant.fields);
             quote! {
                 #ty::#ident #pat => #display
             }
         });
+        let arms = arms.collect::<Vec<_>>();
+        let display_where_clause = display_inferred_bounds.augment_where_clause(input.generics);
         Some(quote! {
             #[allow(unused_qualifications)]
-            impl #impl_generics std::fmt::Display for #ty #ty_generics #where_clause {
+            impl #impl_generics std::fmt::Display for #ty #ty_generics #display_where_clause {
                 fn fmt(&self, __formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
                     #use_as_display
-                    #[allow(
-                        unused_variables,
-                        deprecated,
-                        // Clippy bug: https://github.com/rust-lang/rust-clippy/issues/7422
-                        clippy::nonstandard_macro_braces,
-                        clippy::used_underscore_binding,
-                    )]
+                    #[allow(unused_variables, deprecated, clippy::used_underscore_binding)]
                     match #void_deref self {
                         #(#arms,)*
                     }
@@ -351,7 +422,7 @@ fn impl_enum(input: Enum) -> TokenStream {
         let from_field = variant.from_field()?;
         let backtrace_field = variant.distinct_backtrace_field();
         let variant = &variant.ident;
-        let from = from_field.ty;
+        let from = unoptional_type(from_field.ty);
         let body = from_initializer(from_field, backtrace_field);
         Some(quote! {
             #[allow(unused_qualifications)]
@@ -365,12 +436,18 @@ fn impl_enum(input: Enum) -> TokenStream {
     });
 
     let error_trait = spanned_error_trait(input.original);
+    if input.generics.type_params().next().is_some() {
+        let self_token = <Token![Self]>::default();
+        error_inferred_bounds.insert(self_token, Trait::Debug);
+        error_inferred_bounds.insert(self_token, Trait::Display);
+    }
+    let error_where_clause = error_inferred_bounds.augment_where_clause(input.generics);
 
     quote! {
         #[allow(unused_qualifications)]
-        impl #impl_generics #error_trait for #ty #ty_generics #where_clause {
+        impl #impl_generics #error_trait for #ty #ty_generics #error_where_clause {
             #source_method
-            #backtrace_method
+            #provide_method
         }
         #display_impl
         #(#from_impls)*
@@ -394,6 +471,11 @@ fn fields_pat(fields: &[Field]) -> TokenStream {
 
 fn from_initializer(from_field: &Field, backtrace_field: Option<&Field>) -> TokenStream {
     let from_member = &from_field.member;
+    let some_source = if type_is_option(from_field.ty) {
+        quote!(std::option::Option::Some(source))
+    } else {
+        quote!(source)
+    };
     let backtrace = backtrace_field.map(|backtrace_field| {
         let backtrace_member = &backtrace_field.member;
         if type_is_option(backtrace_field.ty) {
@@ -407,25 +489,43 @@ fn from_initializer(from_field: &Field, backtrace_field: Option<&Field>) -> Toke
         }
     });
     quote!({
-        #from_member: source,
+        #from_member: #some_source,
         #backtrace
     })
 }
 
 fn type_is_option(ty: &Type) -> bool {
+    type_parameter_of_option(ty).is_some()
+}
+
+fn unoptional_type(ty: &Type) -> TokenStream {
+    let unoptional = type_parameter_of_option(ty).unwrap_or(ty);
+    quote!(#unoptional)
+}
+
+fn type_parameter_of_option(ty: &Type) -> Option<&Type> {
     let path = match ty {
         Type::Path(ty) => &ty.path,
-        _ => return false,
+        _ => return None,
     };
 
     let last = path.segments.last().unwrap();
     if last.ident != "Option" {
-        return false;
+        return None;
     }
 
-    match &last.arguments {
-        PathArguments::AngleBracketed(bracketed) => bracketed.args.len() == 1,
-        _ => false,
+    let bracketed = match &last.arguments {
+        PathArguments::AngleBracketed(bracketed) => bracketed,
+        _ => return None,
+    };
+
+    if bracketed.args.len() != 1 {
+        return None;
+    }
+
+    match &bracketed.args[0] {
+        GenericArgument::Type(arg) => Some(arg),
+        _ => None,
     }
 }
 
